use axum::{
    extract::{Path, Query, State},
    Json,
};
use axum_codec::Codec;
use http::StatusCode;
use serde::Deserialize;
use sqlx::query;
use ulid::Ulid;
use uuid::Uuid;

use super::Pagination;
use crate::{
    app::{resource::Resource, SharedState},
    error::AppError,
};

pub fn resource() -> Resource<SharedState> {
    Resource::named("pastes")
        .index(list_pastes)
        .create(create_paste)
        .show(show_paste)
        .update(update_paste)
        .destroy(delete_paste)
}

#[axum_codec::apply(encode)]
struct Paste {
    id: Ulid,
    title: String,
    content: String,
}

async fn list_pastes(
    State(SharedState { db, .. }): State<SharedState>,
    Query(Pagination { limit, offset }): Query<Pagination>,
) -> Result<Codec<Vec<Paste>>, AppError> {
    Ok(Codec(
        query!(
            "SELECT id, title, content from paste LIMIT $1 OFFSET $2",
            limit,
            offset,
        )
        .fetch_all(&db)
        .await?
        .into_iter()
        .map(|r| Paste {
            id: Ulid::from(r.id),
            title: r.title,
            content: r.content,
        })
        .collect(),
    ))
}

#[axum_codec::apply(decode)]
struct CreatePasteRequestBody {
    title: String,
    content: String,
}

async fn create_paste(
    State(SharedState { db, .. }): State<SharedState>,
    Codec(CreatePasteRequestBody { title, content }): Codec<CreatePasteRequestBody>,
) -> Result<Codec<Paste>, AppError> {
    let id = Ulid::new();

    match query!(
        "INSERT INTO paste (id, title, content) VALUES ($1, $2, $3) ON CONFLICT DO NOTHING",
        Uuid::from(id),
        title,
        content,
    )
    .execute(&db)
    .await?
    .rows_affected()
    {
        1 => Ok(Codec(Paste { id, title, content })),
        0 => Err(AppError::UlidConflict(id)),
        rows => Err(AppError::ImpossibleAffectedRows(rows)),
    }
}

async fn show_paste(
    State(SharedState { db, .. }): State<SharedState>,
    Path(id): Path<Ulid>,
) -> Result<Codec<Paste>, AppError> {
    match query!(
        "SELECT title, content FROM paste WHERE id = $1",
        Uuid::from(id),
    )
    .fetch_optional(&db)
    .await?
    {
        Some(r) => Ok(Codec(Paste {
            id,
            title: r.title,
            content: r.content,
        })),
        None => Err(AppError::PasteNotFound(id)),
    }
}

#[derive(Deserialize)]
struct UpdatePasteRequestBody {
    title: String,
    content: String,
}

async fn update_paste(
    State(SharedState { db, .. }): State<SharedState>,
    Path(id): Path<Ulid>,
    Json(UpdatePasteRequestBody { title, content }): Json<UpdatePasteRequestBody>,
) -> Result<StatusCode, AppError> {
    match query!(
        "UPDATE paste SET title = $2, content = $3 WHERE id = $1",
        Uuid::from(id),
        title,
        content,
    )
    .execute(&db)
    .await?
    .rows_affected()
    {
        1 => Ok(StatusCode::NO_CONTENT),
        0 => Err(AppError::PasteNotFound(id)),
        rows => Err(AppError::ImpossibleAffectedRows(rows)),
    }
}

async fn delete_paste(
    State(SharedState { db, .. }): State<SharedState>,
    Path(id): Path<Ulid>,
) -> Result<StatusCode, AppError> {
    match query!("DELETE FROM paste WHERE id = $1", Uuid::from(id))
        .execute(&db)
        .await?
        .rows_affected()
    {
        1 => Ok(StatusCode::NO_CONTENT),
        0 => Err(AppError::PasteNotFound(id)),
        rows => Err(AppError::ImpossibleAffectedRows(rows)),
    }
}
