mod api;
mod pages;
mod resource;
mod root;

use std::sync::Arc;

use axum::{body::Body, Router};
use http::Request;
use memory_serve::{load_assets, MemoryServe};
use sqlx::{postgres::PgConnectOptions, PgPool};
use tower_http::trace::{DefaultOnRequest, DefaultOnResponse, TraceLayer};
use tracing::{field, span, Level};

use crate::config::Config;

#[derive(Clone)]
struct SharedState {
    db: PgPool,
    _config: Arc<Config>,
}

pub async fn build_app(config: Config) -> eyre::Result<Router> {
    let db = PgPool::connect_with(
        PgConnectOptions::new()
            .application_name("ncpn")
            .socket(&config.db_socket)
            .username(&config.db_username)
            .database(&config.db_database),
    )
    .await?;

    let memory_router = MemoryServe::new(load_assets!("./static/")).into_router();

    Ok(Router::new()
        .merge(memory_router)
        .merge(pages::router())
        .nest("/api", api::router())
        .merge(root::router())
        .with_state(SharedState {
            db,
            _config: Arc::new(config),
        })
        .layer(
            TraceLayer::new_for_http()
                .make_span_with(|request: &Request<Body>| {
                    span!(
                        Level::DEBUG,
                        "http-request",
                        uri = field::display(request.uri()),
                    )
                })
                .on_request(DefaultOnRequest::new())
                .on_response(DefaultOnResponse::new()),
        ))
}
