use askama::Template;
use axum::{
    extract::{Query, State},
    response::Redirect,
    routing::get,
    Form, Router,
};
use chrono::{DateTime, TimeZone, Utc};
use maud::{html, Markup};
use serde::Deserialize;
use sqlx::{query, query_as};
use ulid::Ulid;
use uuid::Uuid;

use super::BC_ADMIN;
use crate::{
    app::{
        pages::{dt_iso, page, Breadcrumb, BC_INDEX},
        SharedState,
    },
    error::AppError,
};

pub(super) fn router() -> Router<SharedState> {
    Router::new()
        .route("/", get(show_pastes).post(create_paste))
        .route("/new", get(show_create_paste))
}

const BC_PASTES: Breadcrumb = Breadcrumb::new_static("Pastes", "/admin/pastes");

// async fn show_pastes(
//     State(SharedState { db, .. }): State<SharedState>,
// ) -> Result<Markup, AppError> {
//     Ok(page(
//         "Pastes",
//         &[BC_INDEX, BC_ADMIN],
//         html! {
//             table class="w-full" {
//                 thead {
//                     tr {
//                         th { "Name" }
//                         th { "Created at" }
//                         th { "Actions" }
//                     }
//                 }
//                 tbody {
//                     @for p in query!("SELECT id, title FROM
// paste").fetch_all(&db).await? {                         tr {
//                             td { a.tl href=(format!("/p/{}",
// Ulid::from(p.id))) { (p.title)} }                             td {
// (dt_iso(Ulid::from(p.id).datetime())) }                             td
// class="w-min-content" {                                 div class="flex
// gap-2" {                                     a.tl href="#" { "Edit" }
//                                     a.tl href="#" { "Delete" }
//                                 }
//                             }
//                         }
//                     }
//                 }
//             }
//         },
//     ))
// }

struct PasteInList {
    id: Ulid,
    title: String,
    content: String,
    timestamp: String,
}

#[derive(Template)]
#[template(path = "admin/pastes.html")]
struct ShowPastesTemplate {
    pastes: Vec<PasteInList>,
}

async fn show_pastes(
    State(SharedState { db, .. }): State<SharedState>,
) -> Result<ShowPastesTemplate, AppError> {
    let pastes = query!("SELECT id, title, content FROM paste")
        .fetch_all(&db)
        .await?;
    let pastes = pastes
        .into_iter()
        .map(|r| PasteInList {
            id: r.id.into(),
            title: r.title,
            content: r.content,
            timestamp: DateTime::from_timestamp_millis(Ulid::from(r.id).timestamp_ms() as i64)
                .map(|d| d.to_rfc3339())
                .unwrap_or("<invalid timestamp>".to_string()),
        })
        .collect();
    Ok(ShowPastesTemplate { pastes })
}

#[derive(Template)]
#[template(path = "admin/paste-new.html")]
struct CreatePasteTemplate {
    paste_title: String,
    paste_content: String,
}

#[derive(Deserialize)]
struct CreatePasteFieldsOptional {
    title: Option<String>,
    content: Option<String>,
}

async fn show_create_paste(
    Query(CreatePasteFieldsOptional { title, content }): Query<CreatePasteFieldsOptional>,
) -> CreatePasteTemplate {
    CreatePasteTemplate {
        paste_title: title.unwrap_or_default(),
        paste_content: content.unwrap_or_default(),
    }
}

#[derive(Deserialize)]
struct CreatePasteFields {
    title: String,
    content: String,
}

async fn create_paste(
    State(SharedState { db, .. }): State<SharedState>,
    Form(CreatePasteFields { title, content }): Form<CreatePasteFields>,
) -> Result<Redirect, AppError> {
    let id = Ulid::new();
    match query!(
        "INSERT INTO paste (id, title, content) VALUES ($1, $2, $3)",
        Uuid::from(id),
        title,
        content,
    )
    .execute(&db)
    .await?
    .rows_affected()
    {
        1 => Ok(Redirect::temporary(&format!("/p/{id}"))),
        r => Err(AppError::ImpossibleAffectedRows(r)),
    }
}
