use axum::{routing::MethodRouter, Router};
use axum_codec::{
    handler::Input,
    routing::{delete, get, post, put},
    CodecHandler, IntoCodecResponse,
};

/// A resource which defines a set of conventional CRUD routes.
#[derive(Debug)]
#[must_use]
pub struct Resource<S = ()> {
    pub(crate) name: String,
    pub(crate) router: Router<S>,
}

impl<S> Resource<S>
where
    S: Clone + Send + Sync + 'static,
{
    /// Create a `Resource` with the given name.
    ///
    /// All routes will be nested at `/{resource_name}`.
    pub fn named(resource_name: &str) -> Self {
        Self {
            name: resource_name.to_owned(),
            router: Router::new(),
        }
    }

    /// Add a handler at `GET /{resource_name}`.
    pub fn index<H, I, D, T>(self, handler: H) -> Self
    where
        H: CodecHandler<T, I, D, S> + Clone + Send + Sync + 'static,
        I: Input + Send + 'static,
        D: IntoCodecResponse + Send + Sync + 'static,
        S: Clone + Send + Sync + 'static,
        T: 'static,
    {
        let path = self.index_create_path();
        self.route(&path, get(handler).into())
    }

    /// Add a handler at `POST /{resource_name}`.
    pub fn create<H, I, D, T>(self, handler: H) -> Self
    where
        H: CodecHandler<T, I, D, S> + Clone + Send + Sync + 'static,
        I: Input + Send + 'static,
        D: IntoCodecResponse + Send + Sync + 'static,
        S: Clone + Send + Sync + 'static,
        T: 'static,
    {
        let path = self.index_create_path();
        self.route(&path, post(handler).into())
    }

    /// Add a handler at `GET /{resource_name}/new`.
    pub fn new<H, I, D, T>(self, handler: H) -> Self
    where
        H: CodecHandler<T, I, D, S> + Clone + Send + Sync + 'static,
        I: Input + Send + 'static,
        D: IntoCodecResponse + Send + Sync + 'static,
        S: Clone + Send + Sync + 'static,
        T: 'static,
    {
        let path = format!("/{}/new", self.name);
        self.route(&path, get(handler).into())
    }

    /// Add a handler at `GET /{resource_name}/:{resource_name}_id`.
    pub fn show<H, I, D, T>(self, handler: H) -> Self
    where
        H: CodecHandler<T, I, D, S> + Clone + Send + Sync + 'static,
        I: Input + Send + 'static,
        D: IntoCodecResponse + Send + Sync + 'static,
        S: Clone + Send + Sync + 'static,
        T: 'static,
    {
        let path = self.show_update_destroy_path();
        self.route(&path, get(handler).into())
    }

    /// Add a handler at `GET /{resource_name}/:{resource_name}_id/edit`.
    pub fn edit<H, I, D, T>(self, handler: H) -> Self
    where
        H: CodecHandler<T, I, D, S> + Clone + Send + Sync + 'static,
        I: Input + Send + 'static,
        D: IntoCodecResponse + Send + Sync + 'static,
        S: Clone + Send + Sync + 'static,
        T: 'static,
    {
        let path = format!("/{0}/:{0}_id/edit", self.name);
        self.route(&path, get(handler).into())
    }

    /// Add a handler at `PUT or PATCH /resource_name/:{resource_name}_id`.
    pub fn update<H, I, D, T>(self, handler: H) -> Self
    where
        H: CodecHandler<T, I, D, S> + Clone + Send + Sync + 'static,
        I: Input + Send + 'static,
        D: IntoCodecResponse + Send + Sync + 'static,
        S: Clone + Send + Sync + 'static,
        T: 'static,
    {
        let path = self.show_update_destroy_path();
        // it's 12 AM and I don't know how to get `MethodFilter` to work
        self.route(&path, put(handler.clone()).patch(handler).into())
    }

    /// Add a handler at `DELETE /{resource_name}/:{resource_name}_id`.
    pub fn destroy<H, I, D, T>(self, handler: H) -> Self
    where
        H: CodecHandler<T, I, D, S> + Clone + Send + Sync + 'static,
        I: Input + Send + 'static,
        D: IntoCodecResponse + Send + Sync + 'static,
        S: Clone + Send + Sync + 'static,
        T: 'static,
    {
        let path = self.show_update_destroy_path();
        self.route(&path, delete(handler).into())
    }

    fn index_create_path(&self) -> String {
        format!("/{}", self.name)
    }

    fn show_update_destroy_path(&self) -> String {
        format!("/{0}/:{0}_id", self.name)
    }

    fn route(mut self, path: &str, method_router: MethodRouter<S>) -> Self {
        self.router = self.router.route(path, method_router);
        self
    }
}

impl<S> From<Resource<S>> for Router<S> {
    fn from(resource: Resource<S>) -> Self {
        resource.router
    }
}
